#############################################################################
#                                                                           #
#   Copyright 2019 CERG                                                     #
#                                                                           #
#   Licensed under the Apache License, Version 2.0 (the "License");         #
#   you may not use this file except in compliance with the License.        #
#   You may obtain a copy of the License at                                 #
#                                                                           #
#       http://www.apache.org/licenses/LICENSE-2.0                          #
#                                                                           #
#   Unless required by applicable law or agreed to in writing, software     #
#   distributed under the License is distributed on an "AS IS" BASIS,       #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.#
#   See the License for the specific language governing permissions and     #
#   limitations under the License.                                          #
#                                                                           #
#############################################################################
"""
BIKE-SparseMultSeconWord Power Model
Generates hypothetical power for the RacingBIKE implementation.

The power model is for the second word since we can calculate 
the HD of the first and second word. 
"""
from bitarray import bitarray
from bitarray.util import ba2int, hex2ba, ba2hex, count_xor, int2ba
import numpy as np
import math 
import time

#### Parameters Level 1
# r = 12323
# w = 71
#### Parameters Level 3
# r = 24659
# w = 206//2
#### Parameters Level 5
# r = 40973
# w = 274//2

# r_bits = math.ceil(r/32)*32
# R_WORDS = math.ceil(r/32)

def new_mult_model_opt(dense, sparse, target_round, splice_depth, known_offset, r):
    """
    Input:
        dense: Dense polynomial. (Array of R bits)
        sparse: Indicies of spare polynomial non-zeroe entries. Array of w indicies
        
    Output:
        Hamming distants between words for target round
    """

    R     = r
    R_WDS = (R+31)//32
    
    result_bitarray = bitarray(R)
    
    for rr in range(target_round): 
        # Perform the shift+xor operation
        sparse_idx      = sparse[rr]

        ## Apply rotation and pad to wordlen
        dense_bit_rot = (dense >> sparse_idx) | (dense << (R-sparse_idx))
        dense_bit_rot = dense_bit_rot[0:R]

        result_bitarray = result_bitarray ^ dense_bit_rot

    known_rot = known_offset[-1]

    sparse_idx_bits = sparse[-1]%32
    sparse_idx_wds  = sparse[-1]//32

    guess_rot = sparse_idx_bits+known_rot

    dense_bit_rot = (dense >> guess_rot) | (dense << (R-guess_rot))
    dense_bit_rot = dense_bit_rot[0:R]

    result_bitarray = result_bitarray ^ dense_bit_rot

    hds_a = [None]*splice_depth
    idx   = (sparse_idx_wds)%R_WDS
    idxp1 = (sparse_idx_wds+1)%R_WDS

    for i in range(32-(R%32)):
        result_bitarray.append(0)

    for i in range(0,splice_depth):
        # if (ba2hex(result_bitarray[32*idx:32*(idx+1)]) == "b7d7ef51"):
        #     print(sparse[-1], ba2hex(result_bitarray[32*idx:32*(idx+1)]), ba2hex(result_bitarray[32*idxp1:32*(idxp1+1)]), " -> ", count_xor(result_bitarray[32*idx:32*(idx+1)], result_bitarray[32*idxp1:32*(idxp1+1)]))
        hds_a[i] = count_xor(result_bitarray[32*idx:32*(idx+1)], result_bitarray[32*idxp1:32*(idxp1+1)])
        idx   = (idx + 1) % R_WDS
        idxp1 = (idx + 1) % R_WDS
    return hds_a

def new_mult_model_opt_alt(dense, sparse, target_round, splice_depth, r):
    """
    Input:
        dense: Dense polynomial. (Array of R bits)
        sparse: Indicies of spare polynomial non-zeroe entries. Array of w indicies
        
    Output:
        Hamming distants between words for target round
    """

    R     = r
    R_WDS = (R+31)//32
    
    result_bitarray = bitarray(R)
    
    for rr in range(target_round+1): 
        # Perform the shift+xor operation
        sparse_idx      = sparse[rr]

        ## Apply rotation and pad to wordlen
        dense_bit_rot = (dense >> sparse_idx) | (dense << (R-sparse_idx))
        dense_bit_rot = dense_bit_rot[0:R]

        result_bitarray = result_bitarray ^ dense_bit_rot

    hds_a = [None]*splice_depth
    sparse_idx_wds  = sparse[-1]//32
    tgt_offset = (384-sparse_idx_wds)

    # idx   = (sparse_idx_wds+tgt_offset)%R_WDS
    # idxp1 = (sparse_idx_wds+1+tgt_offset)%R_WDS

    idx   = (sparse_idx_wds)%R_WDS
    idxp1 = (sparse_idx_wds+1)%R_WDS

    for i in range(32-(R%32)):
        result_bitarray.append(0)

    for i in range(0,splice_depth):
        # print(sparse[-1], ba2hex(result_bitarray[32*idx:32*(idx+1)]), ba2hex(result_bitarray[32*idxp1:32*(idxp1+1)]))
        hds_a[i] = count_xor(result_bitarray[32*idx:32*(idx+1)], result_bitarray[32*idxp1:32*(idxp1+1)])
        idx   = (idx + 1) % R_WDS
        idxp1 = (idx + 1) % R_WDS
    # print()
    return hds_a


####################
# HELPER FUNCITONS #
####################
def parse_bike_fobos_tv_vecline(tv_str, r_param):    
    r_bits = math.ceil(r_param/32)*32

    if (r_param == 12323):
        w = 71
    elif (r_param == 24659):
        w = 206//2
    else:
        w = 274//2

    start_cmd = "00C0"
    len_cmd_chars = 4
    
    ## Find where PDI in starts
    data_start = tv_str.index(start_cmd) + len(start_cmd) + len_cmd_chars + 32//4
    
    ## H0 is the first data input from the PDI string
    h0_str = tv_str[data_start:data_start+w*8]
    
    ## C0 is located after H0, H1, and SIGMA
    h1_str_start = data_start+w*8
    c0_start = h1_str_start + w*8 + 256//4
    c0_str = tv_str[c0_start:c0_start+(r_bits+3)//4]

    h0_list = [int(h0_str[i:i+8],16) for i in range(0, len(h0_str), 8)]

    # word order of C0 is big endian - need to flip
    c0_list = [c0_str[i:i+8] for i in range(0, len(c0_str), 8)]
    for i,word in enumerate(c0_list):
        c0_list[i] = "".join([word[i:i+2] for i in range(0, len(word), 2)][::-1])
    c0_list = "".join(c0_list)

    c0_list = [c0_list[i:i+2] for i in range(0, len(c0_str), 2)]

    end = (4-((r_param//2)%4))%4
    # c0_list = c0_list[:-end]

    # Parse into bit array
    c0_bitarray = bitarray()

    for i in range(r_param):
        if ((i % 8) == 0):
            b = int(c0_list[i//8],16) 
        c0_bitarray.append(((b >> (i % 8)) & 0x1))    


    return h0_list, c0_bitarray

def getHypotheticalPower(dinFile, numTraces, splice_depth, mult_target_round, known_keys, known_offset, r_param):
    # Generate powermodel for first round where we target the lowest 
    # log2(b) bits of the first index of h0. HD of the first and 
    # second word written to memory.
    
    # parameters:
    # -------
    # plaintextFile: string
    #     Plain text file name. The file contains plaintext in Hex format
    #     with space seperating bytes. The file will be loaded using
    #     np.loadtxt
    # ciphertextFile: string
    #     Cipher text file name. The file contains ciphertext in Hex format
    #     with space seperating bytes The file will be loaded using
    #     np.loadtxt
    # numTraces:
    #     The number of encryption operations performed.

    R_WORDS = math.ceil(r_param/32)
    
    print('---- Loading powermodel input data:')
    publicinput, secretinput = loadInput(dinFile, numTraces, r_param)

    print("First word of h0={}".format(secretinput[0][0]))
    actual_key = secretinput[0][0]
    print('---- Loading powermodel input data complete.')
    print('---- Calculating hypothetical power for all subkeys:')
    # hypotheticalPower = []
    # hypotheticalPower = np.array()
    start_t = time.time()

    subkey_space = r_param # <- debug runtime
    offset       = 0
    step         = 1

    keys = list(range(offset,offset+subkey_space, step))

    for ii in range(numTraces):    
        if (actual_key != secretinput[ii][0]):
            print("WARNING: Key is not constant. Expected h0[0]={}, found h0[0]={}".format(actual_key, secretinput[ii][0]))
        
        # Generate HD model for the splice_depth writes of the target
        layer_hds    = np.zeros((len(keys),splice_depth))
        for k in keys:
            if (k not in known_keys):
                if (len(known_keys) == 0 or len(known_offset) == 0):
                    layer_hds[keys.index(k),:] = new_mult_model_opt_alt(publicinput[ii], known_keys+[k], mult_target_round, splice_depth, r_param)
                else:
                    layer_hds[keys.index(k),:] = new_mult_model_opt(publicinput[ii], known_keys+[k], mult_target_round, splice_depth, known_offset, r_param)   
            else:
                layer_hds[keys.index(k),:] = [0]*splice_depth
                layer_hds[keys.index(k),0] = 1


        # Parse into individual key guess power models
        if ii == 0:
            hypotheticalPower = layer_hds.transpose()
        else:
            hypotheticalPower = np.concatenate([hypotheticalPower, layer_hds.transpose()], axis=0)


        # for i in range(min([splice_depth,R_WORDS-1])):
        #     oneWordPower = layer_hds[:,i]
        #     hypotheticalPower.append(oneWordPower)

        if (ii % 10 == 0):
            print("i:",ii)
    
    print('---- Hypothetical power calculation done.')
    end_t = time.time()
    print(end_t - start_t)

    # return hypotheticalPower
    return hypotheticalPower





# Functions to be vectorized
def wordGetHD(D1, D2):
    Ddelta = D1 ^ D2
    
    hd = 0
    for i in range(32):
        if (Ddelta >> i & 0x1 == 1):
            hd += 1

    return hd


#################################################################

def loadInput(filename, num, r_param):
    h0_vec = []
    c0_vec = []
    
    with open(filename, "r") as f:
        for i,line in enumerate(f):
            if (i == num): 
                break
            h0,c0 = parse_bike_fobos_tv_vecline(line, r_param)
            
            h0_vec.append(h0)
            c0_vec.append(c0)     

    return c0_vec, h0_vec

